/*====================================================================================
This file is part of the TechTools Plugin Development Kit. 
Copyright (c) 2011 by TechTools  

You may modify this file and/or create derived works to develop plugins 
for use with TechTools's DigiView software and capture files obtained from
TechTools' hardware. Any other use is prohibited.

email: support@tech-tools.com,  sales@tech-tools.com 
web: www.tech-tools.com   Voice:972-272-9392  FAX: 972-494-5814
======================================================================================*/


#include "plugin.h"
// CmdParser.cpp  V0.1 BETA  Copyright (c) 2011 by TechTools

// You do not need to (and should not) make any changes in this file.  ALL of your code
// goes in your project-specific source file.  
// We include this source for reference.

// We might make changes to this file to support new features in the future. 
// If you avoid making any changes here, you might be able to recompile your 
// plugin with the new version and maintain compatibility.

void StartFrame(int64 timestamp, Data64 data, unsigned char FormatID)
{
	data.bytes[7] = 0x86;           // frame+data
	data.bytes[6] = FormatID;
	fwrite(data.bytes,8,1,stdout);  // send the data
	data.i64 = timestamp;           //   and now 
	fwrite(data.bytes,8,1,stdout);  // the timestamp	
}

void StartField(int64 timestamp, Data64 data, unsigned char FormatID)
{
	data.bytes[7] = 0x84;           // data .. no frame
	data.bytes[6] = FormatID;
	fwrite(data.bytes,8,1,stdout);  // send the data
	data.i64 = timestamp;           //   and now 
	fwrite(data.bytes,8,1,stdout);  // the timestamp	
}

void EndFrame(int64 timestamp)
{
	Data64 data;
	data.i64 = 0;
	data.bytes[7] = 0x87;           // frame idle
	fwrite(data.bytes,8,1,stdout);  // send the data
	data.i64 = timestamp;           //   and now 
	fwrite(data.bytes,8,1,stdout);  // the timestamp	
}

void EndField(int64 timestamp)
{
	Data64 data;
	data.i64 = 0;
	data.bytes[7] = 0x85;           // data idle
	fwrite(data.bytes,8,1,stdout);  // send the data
	data.i64 = timestamp;           //   and now 
	fwrite(data.bytes,8,1,stdout);  // the timestamp	
}


void SendControl(int64 timestamp,unsigned char CntlCode)
{
	Data64 data;
	if (CntlCode < CNTLUNKNOWN)
	{
		data.i64 = 0;
		data.bytes[7] = 0x80;           // is a control field
		data.bytes[6] = CntlCode;       
		fwrite(data.bytes,8,1,stdout);  // send the data
		data.i64 = timestamp;           //   and now 
		fwrite(data.bytes,8,1,stdout);  // the timestamp	
	}
}


void FindChannelLimits(uint64 mask, uint64 &HighestBit, uint64 &LowestBit)
{
	HighestBit = 0x8000000000000000;
	LowestBit = 1;

	if (mask != 0)
	{
		while ((mask & HighestBit) == 0) 
			HighestBit >>= 1;

		while ((mask & LowestBit) == 0) 
			LowestBit <<= 1;
	}
}


/*Extracts the data for a given signal from the rawdata.
The selected bits are packed together and '0' bit justified.
MASK should be the bit mask for a given signal (from a chanselect cfg).
HighBit and LowBit are the highest and lowest channels used
by THIS signal (prev. determined with 'FindChannelLimits'.
*/
uint64 pack(uint64 dat, uint64 mask, uint64 HighBit, uint64 LowBit)
{
	uint64 res;
	res = 0;
	if ((dat & mask) != 0)
	{
		LowBit >>= 1;
		do  // scan from the Highest bit to the Lowest bit used by this signal
		{ 
			if ((mask & HighBit) != 0)     
			{ 
				res <<= 1;                  // include this data bit
				if ((dat & HighBit) != 0) // set LSB to the data bit value
					res |= 1;
			}
			HighBit >>= 1;  // check to see if the next bit is part of this signal
			if (HighBit == 0) break;  // incase LowBit is illegal (> Highbit or multi-bits set)
		} while (HighBit != LowBit);
	}

	return res;
}


// private routine to send a vector of strings back to the app.
static void WriteStrings(vector<string> &slist) 
{
	unsigned int i;
	for (i=0;i<slist.size();i++)
	{    
		if (slist[i].length() < 1)  // don't allow empty string to terminate list.
			slist[i] = "<empty>";   // Give parser something descriptive to print in error messages.
		fwrite( slist[i].c_str(), slist[i].length()+1,1,stdout);
	}

	fwrite("\0",1,1,stdout);  // terminate the list
}


// main command parsing loop.
int _tmain(int argc, _TCHAR* argv[])
{
	enum mycmds { GET_STR, SET_CFG, GO_CMD, EXIT_CMD, STOP_CMD, INIT_CMD  };

	Data64 datin;
	static Data64 rawdata;

	// CIN and COUT are clean and versatile, but slow.  
	// We will use the block oriented fwrite,fread and fflush
	// on the underlying stdin/stdout handles for performance.  But first...

	// ensure we are using adequate buffering
	setvbuf(stdout, NULL, _IOFBF, 4096);
	setvbuf(stdin, NULL, _IOFBF, 4096);

	// set STDIO to binary mode
	_setmode(_fileno(stdin), _O_BINARY);
	_setmode(_fileno(stdout), _O_BINARY);

	vector<string> strlist;

	// Let the plug-in do any start up tasks
	OnLoad();

	// fetch data packets from the host app until stream error or commanded to exit
	while (fread(datin.bytes,1,8,stdin) == 8)  // fetch 8 bytes of data, returns FALSE on stream close or error
	{   
		if (datin.bytes[7] == (0x40 + EXIT_CMD))  // done
			break;

		if ((datin.bytes[7] & 0xC0) == 0x40)   // command 
		{
			switch (datin.bytes[7] & 0x3F)
			{

			case GET_STR:    // get a string descriptor ([6] selects which one)
				fwrite(datin.bytes,8,1,stdout);    // echo the full command packet...
				strlist.clear();
				GetStrList(datin.bytes[6],strlist);
				WriteStrings(strlist);
				fflush(stdout);
				break;

			case SET_CFG: // set one configuration item
				fwrite(datin.bytes,8,1,stdout);    // echo the full command packet...
				SetCfgItem(datin.bytes[6], datin.bytes[5], datin.lowint);
				break;

			case GO_CMD:        // START parsing protocol...data to follow
				fwrite(datin.bytes,8,1,stdout);    // echo the full command packet...
				StartOfData();  // let protocol parser initialize
				//	response ignored until streaming, don't flush
				break;

			case STOP_CMD:     // End-of-Data/STOP parsing protocol
				EndOfData();   // let protocol parser clean up
				fwrite(datin.bytes,8,1,stdout);    // echo the full command packet...
				fflush(stdout);
				break;

			case INIT_CMD: // set one init item
				fwrite(datin.bytes,8,1,stdout);    // echo the full command packet...
				SetInitItem(datin.bytes[6], datin.bytes[5], datin.lowint);
				break;

			default:    // unknown command
				fwrite(datin.bytes,8,1,stdout);    // echo the full command packet...
				fflush(stdout);
				break;

			}  // command switch
		}  // is a command

		else if ((datin.bytes[7] & 0xC0) == 0x80)   // is DATA
			rawdata = datin;   // remember the data - send with timestamp later

		else // 0 or C => is timestamp
			Parse(datin.i64,rawdata);  // send the sample data and time to the plugin's parser()

	}  // until terminate command or stream error

	// Let the plug-in do any final clean up tasks
	OnUnLoad();
	fflush(stdout);

	return 0;
}

